package quimufu.colourful_portals.entity;

import net.minecraft.block.BlockState;
import net.minecraft.block.Blocks;
import net.minecraft.client.world.ClientWorld;
import net.minecraft.entity.Entity;
import net.minecraft.entity.EntityType;
import net.minecraft.entity.LivingEntity;
import net.minecraft.entity.projectile.thrown.ThrownItemEntity;
import net.minecraft.item.Item;
import net.minecraft.particle.ParticleTypes;
import net.minecraft.predicate.entity.EntityPredicates;
import net.minecraft.registry.Registries;
import net.minecraft.registry.RegistryKey;
import net.minecraft.registry.RegistryKeys;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.network.ServerPlayerEntity;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.sound.SoundCategory;
import net.minecraft.sound.SoundEvents;
import net.minecraft.util.Identifier;
import net.minecraft.util.hit.HitResult;
import net.minecraft.util.math.*;
import net.minecraft.world.Heightmap;
import net.minecraft.world.TeleportTarget;
import net.minecraft.world.World;
import net.minecraft.world.border.WorldBorder;
import net.minecraft.world.dimension.DimensionType;
import org.jetbrains.annotations.Nullable;
import org.joml.Vector3f;
import quimufu.colourful_portals.ColourfulPortalsMod;
import quimufu.colourful_portals.config.ColourfulPortalConfig;
import quimufu.colourful_portals.portal.PortalHelper;
import quimufu.colourful_portals.util.AdditionalMath;

import java.util.List;

public class ColourfulPearlEntity
        extends ThrownItemEntity {

    public ColourfulPearlEntity(EntityType<? extends ColourfulPearlEntity> entityType, World world) {
        super(entityType, world);
    }

    public ColourfulPearlEntity(World world, LivingEntity owner) {
        super(ColourfulPortalsMod.COLOURFUL_PEARL_ENTITY_TYPE, owner, world);
    }


    @Override
    protected Item getDefaultItem() {
        return ColourfulPortalsMod.COLOURFUL_PEARL_ITEM;
    }

    @Override
    public void onRemoved() {
        if (this.getRemovalReason() == RemovalReason.DISCARDED) {
            if (getWorld() instanceof ClientWorld clientWorld) {
                clientWorld.playSound(
                        this.getX(), this.getY(), this.getZ(),
                        ColourfulPortalsMod.TELEPORT_AWAY_SOUND,
                        SoundCategory.BLOCKS,
                        100F, 1.25F,
                        false);
            }

            List<Entity> entities = getWorld()
                    .getOtherEntities(this, this.getBoundingBox().expand(5));

            entities.stream()
                    .filter(e -> e.squaredDistanceTo(this) < 25)
                    .forEach(this::spawnTeleportationParticles);
        }
    }

    private void spawnTeleportationParticles(Entity entity) {
        for (int i = 0; i < 64; ++i) {
            Box box = entity.getBoundingBox();
            float xOffset = (float) ((AdditionalMath.root(random.nextFloat() * 2 - 1, 3) / 2) * box.getLengthX() * 2);
            float yOffset = (float) (AdditionalMath.root(random.nextFloat(), 3) * box.getLengthY() * 2);
            float zOffset = (float) ((AdditionalMath.root(random.nextFloat() * 2 - 1, 3) / 2) * box.getLengthZ() * 2);

            Vector3f offset = new Vector3f(xOffset, yOffset, zOffset);
            addHitParticle(this.getWorld(), entity, offset);
        }
    }

    @Override
    protected void onCollision(HitResult hitResult) {
        if (getWorld().isClient()) {
            return;
        }
        if (isRemoved()) {
            return;
        }
        List<Entity> entities = getWorld()
                .getOtherEntities(this, this.getBoundingBox().expand(5), EntityPredicates.EXCEPT_SPECTATOR
                        .and(entity -> entity instanceof LivingEntity)
                        .and(entity -> !entity.getType().isIn(ColourfulPortalsMod.COLOURFUL_PEARL_NOT_TELEPORTABLE)));
        entities.stream()
                .filter(e -> e.squaredDistanceTo(this) < 25)
                .forEach(this::tryTeleport);
        this.discard();
    }

    private void tryTeleport(Entity entity) {
        if (!(getWorld() instanceof ServerWorld serverWorld)) {
            return;
        }
        if (entity == null || !canTeleportEntityTo(entity, serverWorld)) {
            return;
        }
        if (entity.hasVehicle()) {
            entity.detach();
        }
        TeleportTarget target = findTarget(serverWorld, entity, null);
        while (target == null || !targetValid(target, entity)) {
            target = findTarget(serverWorld, entity, target);
        }
        prepareTarget(target, entity);

        if (entity instanceof ServerPlayerEntity serverPlayerEntity) {
            if (serverPlayerEntity.networkHandler.isConnectionOpen()) {

                Entity entityAfterTeleport = entity.teleportTo(target);
                if (entityAfterTeleport == null) {
                    return;
                }
                entityAfterTeleport.onLanding();
                serverPlayerEntity.clearCurrentExplosion();
                entityAfterTeleport.damage(this.getDamageSources().fall(), 5.0f);
                serverPlayerEntity.playSoundToPlayer(SoundEvents.ENTITY_PLAYER_TELEPORT, SoundCategory.PLAYERS, 1F, 0.75F);
            }
        } else {
            Entity entityAfterTeleport = entity.teleportTo(target);
            if (entityAfterTeleport == null) {
                return;
            }
            entityAfterTeleport.onLanding();
            serverWorld.playSound(entityAfterTeleport, entityAfterTeleport.getBlockPos(),
                    SoundEvents.ENTITY_PLAYER_TELEPORT, SoundCategory.NEUTRAL,
                    1F, 0.75F);
        }
    }

    private void addHitParticle(World world, Entity entity, Vector3f offset) {
        Vector3f endPos = randomInside(this.getDimensions(null).getBoxAt(entity.getPos()))
                .add(offset);
        world.addParticle(ParticleTypes.PORTAL,
                endPos.x, endPos.y, endPos.z,
                -offset.x, -offset.y, -offset.z);
    }

    private Vector3f randomInside(Box box) {
        return box.getMinPos().toVector3f()
                .add(((float) box.getLengthX()) * random.nextFloat(),
                        ((float) box.getLengthY()) * random.nextFloat(),
                        ((float) box.getLengthZ()) * random.nextFloat());
    }

    private boolean targetValid(TeleportTarget target, Entity entity) {
        Box boundingBox = entity.getDimensions(entity.getPose())
                .getBoxAt(target.pos())
                .stretch(0, -1, 0);
        boolean foundAir = false;
        boolean foundNonAir = false;
        for (BlockPos pos : PortalHelper.blockPosInBox(boundingBox)) {
            BlockState blockState = target.world().getBlockState(pos);
            if (!blockState.isIn(ColourfulPortalsMod.COLOURFUL_PEARL_REPLACEABLE_BLOCK_TAG)) {
                ColourfulPortalsMod.LOGGER.info("invalid location at {}, block {}", pos, Registries.BLOCK.getId(blockState.getBlock()));
                return false;
            }
            if (blockState.isAir()) {
                foundAir = true;
            } else {
                foundNonAir = true;
            }
        }
        return foundAir && foundNonAir;
    }


    private void prepareTarget(TeleportTarget target, Entity entity) {
        Box entityBody = entity.getDimensions(entity.getPose()).getBoxAt(target.pos());
        double minY = entityBody.getMinPos().getY();
        Box floor = entityBody.withMinY(minY - 1).withMaxY(minY - 0.5);
        for (BlockPos pos : PortalHelper.blockPosInBox(entityBody)) {
            target.world().setBlockState(pos, ColourfulPortalsMod.COLOURFUL_AIR.getRandomState(random));
        }
        for (BlockPos pos : PortalHelper.blockPosInBox(floor)) {
            target.world().setBlockState(pos, Blocks.STONE.getDefaultState());
        }

    }

    private @Nullable TeleportTarget findTarget(ServerWorld serverWorld, Entity entity, TeleportTarget target) {
        if (target != null
            //position makes sense to retry
            && !isEmptyPosition(target)
            //prevents potential infinite loop if position does have weird block compositions
            && random.nextInt(target.world().getLogicalHeight() * 2) != 0) {
            return new TeleportTarget(target.world(),
                    target.pos().withAxis(Direction.Axis.Y, getTargetY(target.world())),
                    target.velocity(),
                    target.yaw(),
                    target.pitch(),
                    target.postDimensionTransition());
        }
        //re-use targetWorld to prevent dimensions with less valid spawning positions from becoming less likely
        ServerWorld targetWorld = target == null ? getTargetWorld(serverWorld) : target.world();
        if (targetWorld == null) {
            return null;
        }
        //adjust position within target
        Vec3d minPos = Vec3d.of(getTargetPos(serverWorld, targetWorld, getPos()));
        int width = MathHelper.ceil(entity.getBoundingBox().getLengthX());
        int depth = MathHelper.ceil(entity.getBoundingBox().getLengthZ());
        Vec3d targetLocation = minPos.add((width % 2) * 0.5F, 0, (depth % 2) * 0.5F);

        return new TeleportTarget(targetWorld, targetLocation, entity.getVelocity(), entity.getYaw(), entity.getPitch(), TeleportTarget.NO_OP);
    }

    private static boolean isEmptyPosition(TeleportTarget target) {
        return target.world().getTopY(Heightmap.Type.WORLD_SURFACE, (int) target.pos().x, (int) target.pos().z) == target.world().getBottomY();
    }

    private @Nullable ServerWorld getTargetWorld(ServerWorld serverWorld) {
        if (this.getRandom().nextDouble() < ColourfulPortalConfig.pearlSameDimensionLikelihood) {
            return serverWorld;
        }

        MinecraftServer server = serverWorld.getServer();
        ColourfulPortalConfig.addMissingDimensionsToConfig(server);
        Identifier targetWorldId = ColourfulPortalsMod.DIMENSION_WEIGHTS_COLOURFUL_PEARL.getWeighted(random);

        return server.getWorld(RegistryKey.of(RegistryKeys.WORLD, targetWorldId));
    }

    private BlockPos getTargetPos(ServerWorld fromWorld, ServerWorld toWorld, Vec3d pos) {

        double distance = ColourfulPortalConfig.minPearlDistance +
                          getRandom().nextDouble() *
                          (ColourfulPortalConfig.maxPearlDistance - ColourfulPortalConfig.minPearlDistance);

        double angle = Math.PI * 2 * getRandom().nextDouble();
        int targetY = getTargetY(toWorld);

        Vec3d target = new Vec3d(pos.getX() + Math.cos(angle) * distance, targetY, pos.getZ() + Math.sin(angle) * distance);

        WorldBorder worldBorder = toWorld.getWorldBorder();
        double d = DimensionType.getCoordinateScaleFactor(fromWorld.getDimension(), toWorld.getDimension());
        return worldBorder.clamp(target.getX() * d, target.getY(), target.getZ() * d);
    }

    private int getTargetY(ServerWorld toWorld) {
        return toWorld.getBottomY() + 4 + (getRandom().nextInt(toWorld.getLogicalHeight() - 8));
    }

    private static boolean canTeleportEntityTo(Entity entity, World world) {
        if (entity.getWorld().getRegistryKey() == world.getRegistryKey()) {
            if (entity instanceof LivingEntity livingEntity) {
                return livingEntity.isAlive() && !livingEntity.isSleeping();
            }
            return entity.isAlive();
        }
        return entity.canUsePortals(true);
    }
}
